theory LTS_Parser
  imports "$AFP/XML/Xmlt" LIA_Instance
begin

hide_const (open) name

(* move? *)

type_synonym 'a xmlst = "xml list \<Rightarrow> string +\<^sub>\<bottom> 'a"

definition xml_if :: "tag \<Rightarrow> 'a xmlt \<Rightarrow> 'a xmlt \<Rightarrow> 'a xmlt" where
  "xml_if tag p1 p2 xml \<equiv>
   case xml of XML nam _ _ \<Rightarrow> if nam = tag then p1 xml else p2 xml
   | _ \<Rightarrow> Xmlt.fail tag xml"

lemma xml_if_mono [partial_function_mono]:
  assumes "\<And>y. mono_sum_bot (p1 y)"
    and "\<And>y. mono_sum_bot (p2 y)"
  shows "mono_sum_bot (\<lambda>g. xml_if t (\<lambda>y. p1 y g) (\<lambda> y. p2 y g) x)"
  using assms
    by (cases x, auto intro!: partial_function_mono simp: xml_if_def)

definition xml_parse_text :: "string xmlt" where
  "xml_parse_text xml \<equiv>
   case xml of XML_text text \<Rightarrow> return text
   | _ \<Rightarrow> error (show ''could not extract text from\<newline>'' @ show xml)"

definition xml_enter :: "tag \<Rightarrow> 'a xmlst \<Rightarrow> 'a xmlt" where
  "xml_enter tag p xml \<equiv>
   case xml of XML nam [] xs \<Rightarrow> if nam = tag then p xs else Xmlt.fail tag xml
   | _ \<Rightarrow> Xmlt.fail tag xml"

definition "xml_enter_if tag ps p2 xml \<equiv>
   case xml of XML nam [] xs \<Rightarrow> if nam = tag then ps xs else p2 xml"

fun xml_enter_options_sub where
  "xml_enter_options_sub tags ((tag,p)#rest) (XML name atts xmls) =
    (if tag = name then p xmls
     else xml_enter_options_sub (name@tags) rest (XML name atts xmls))"
| "xml_enter_options_sub tags _ xml =
    error (''expecting '' @ show (rev tags) @ '' but encountered\<newline>'' @ show xml)"

definition "xml_enter_options = xml_enter_options_sub []"

definition "xml_take_text p xmls \<equiv>
   case xmls of XML_text text # xmls \<Rightarrow> p text xmls
   | _ \<Rightarrow> error (show ''expecting a text but encountered\<newline>'' @ show xmls)"

definition "xml_take_int p xs \<equiv>
  case xs of [] \<Rightarrow> error (''expecting an integer but tag closed'')
  | XML_text text # xs \<Rightarrow>
    (case (int_of_string text) of Right n \<Rightarrow> p n xs | Left e \<Rightarrow> Left e)
  | x#xs \<Rightarrow> error (''expecting an integer but encountered\<newline>'' @ show x)"

text {* parses the first child *}
fun xml_take :: "'a xmlt \<Rightarrow> ('a \<Rightarrow> xml list \<Rightarrow> string +\<^sub>\<bottom> 'b) \<Rightarrow> xml list \<Rightarrow> string +\<^sub>\<bottom> 'b" where
  "xml_take p1 p2 [] = error (''unexpected tag close'')"
| "xml_take p1 p2 (x # xs) = (case p1 x of Right a \<Rightarrow> p2 a xs | Left x \<Rightarrow> Left x)"

fun xml_take_optional :: "tag \<Rightarrow> 'a xmlt \<Rightarrow> ('a option \<Rightarrow> xml list \<Rightarrow> string +\<^sub>\<bottom> 'b) \<Rightarrow> xml list \<Rightarrow> string +\<^sub>\<bottom> 'b" where
  "xml_take_optional tag p1 p2 [] = p2 None []"
| "xml_take_optional tag p1 p2 (x # xs) =
    (case x of XML intag _ _ \<Rightarrow>
      if intag = tag then
        (case p1 x of Right a \<Rightarrow> p2 (Some a) xs | Left x \<Rightarrow> Left x)
      else p2 None (x#xs))"

fun xml_take_all_sub :: "tag \<Rightarrow> nat \<Rightarrow> nat option \<Rightarrow> 'a xmlt \<Rightarrow> 'a list \<Rightarrow> ('a list \<Rightarrow> xml list \<Rightarrow> string +\<^sub>\<bottom> 'b) \<Rightarrow> xml list \<Rightarrow> string +\<^sub>\<bottom> 'b" where
  "xml_take_all_sub tag minOccurs _ p1 acc p2 [] =
    (if minOccurs = 0 then p2 (rev acc) [] else error (''too few '' @ show tag))"
| "xml_take_all_sub tag minOccurs None p1 acc p2 (x # xs) =
    (case x of XML intag inatt inner \<Rightarrow>
      if intag = tag then
        case p1 x of Right a \<Rightarrow>
          xml_take_all_sub tag (minOccurs-1) None p1 (a # acc) p2 xs
        | Left x \<Rightarrow> Left x
      else if minOccurs = 0 then p2 (rev acc) (x # xs)
      else error (''too few '' @ show tag @ '' before '' @ show intag)
     | _ \<Rightarrow> Xmlt.fail tag x)"
| "xml_take_all_sub tag minOccurs (Some maxOccurs) p1 acc p2 (x#xs) =
    (if maxOccurs > 0 then
      (case p1 x of Right a \<Rightarrow>
       xml_take_all_sub tag (minOccurs-1) (Some (maxOccurs-1)) p1 (a#acc) p2 xs
       | _ \<Rightarrow> Xmlt.fail tag x)
     else error (''too many '' @ show tag @ ''\<newline>'' @ show x))"

definition xml_take_all where
  "xml_take_all tag minOccurs maxOccurs p1 p2 xs \<equiv>
   xml_take_all_sub tag minOccurs maxOccurs p1 [] p2 xs"

fun xml_leave :: "'a \<Rightarrow> xml list \<Rightarrow> string +\<^sub>\<bottom> 'a" where
  "xml_leave v [] = Right v"
| "xml_leave v (x#xs) = error (''want to close but found\<newline>'' @ show x)"

fun drop_xml_attributes :: "xml \<Rightarrow> xml" where
  "drop_xml_attributes (XML n atts cs) = (XML n [] cs)"
| "drop_xml_attributes x = x"

context
  fixes atom_parser :: "('f,'a,'t) exp formula xmlt"
begin

partial_function (sum_bot) formula_parser where
  [code]: "formula_parser xml =
     (xml_if ''disjunction'' (Xmlt.many ''disjunction'' formula_parser Disjunction)
     (xml_if ''conjunction'' (Xmlt.many ''conjunction'' formula_parser Conjunction)
     atom_parser)) xml"

end

context
  fixes location_parser :: "'l xmlt"
    and trans_parser :: "'tr xmlt"
    and tatom_parser :: "('f,'v trans_var,'t) exp formula xmlt"
begin

definition transition_parser :: "('tr \<times> ('f,'v,'t,'l) transition_rule) xmlt" where
  "transition_parser \<equiv> Xmlt.tuple4 ''transition''
    trans_parser
    (Xmlt.singleton ''source'' location_parser id)
    (Xmlt.singleton ''target'' location_parser id)
    (Xmlt.singleton ''formula'' (formula_parser tatom_parser) id)
    (\<lambda> tr l r \<phi>. (tr, Transition l r \<phi>))" 

definition lts_parser :: "tag \<Rightarrow> ('f,'v,'t,'l,'tr) lts_impl xmlt" where
  "lts_parser tag \<equiv> Xmlt.many1 tag
     (Xmlt.singleton ''initial'' location_parser id)
     transition_parser (\<lambda> i t. Lts_Impl i t [])"

definition "safety_input_parser \<equiv>
  Xmlt.pair ''ltsSafetyInput'' (lts_parser ''lts'') (Xmlt.many ''error'' location_parser id) Pair"

end

context
  fixes art_node_id_parser :: "tag \<Rightarrow> 'n xmlt"
    and location_parser :: "'l xmlt"
    and trans_parser :: "'tr xmlt"
    and atom_parser :: "('f,'v,'t) exp formula xmlt"
    and hint_parser :: "'h :: default xmlt"
begin

definition hints_parser :: "'h hint xmlt" where
  "hints_parser = Xmlt.many ''hints'' hint_parser (\<lambda>hs. hint.Conjunction (map hint.Base hs))"

definition art_node_parser :: "('f, 'v, 't, 'l, 'n, 'tr, 'h hint) art_node_impl xmlt" where
  "art_node_parser =
     Xmlt.tuple4 ''node''
     (art_node_id_parser ''nodeId'')
     (Xmlt.singleton ''invariant'' (formula_parser atom_parser) id)
     (Xmlt.singleton ''location'' location_parser id)
     (Xmlt.options [
       (''children'', Xmlt.many ''children'' 
         (xml_enter ''child'' (
           xml_take trans_parser (\<lambda>tr.
           xml_take (art_node_id_parser ''nodeId'') (\<lambda>n.
           xml_take_optional ''hints'' hints_parser (\<lambda>h_op.
           xml_leave (tr,n, (case h_op of Some h \<Rightarrow> h | _ \<Rightarrow> default))))))) Children_Edge),
       (''coverEdge'', xml_enter ''coverEdge'' (
          xml_take (art_node_id_parser ''nodeId'') (\<lambda>n.
          xml_take_optional ''hints'' hints_parser (\<lambda>h_op.
          xml_leave (Cover_Edge n (case h_op of Some h \<Rightarrow> h | _ \<Rightarrow> default))))))])
     Art_Node"

definition art_parser :: "(('f, 'v, 't, 'l, 'n, 'tr, 'h hint) art_impl) xmlt" where
  "art_parser = Xmlt.pair ''impact''
     (art_node_id_parser ''initial'')
     (Xmlt.many ''nodes'' art_node_parser id)
     (\<lambda> init ns. (\<lparr> art_impl.initial_node = init, nodes = ns \<rparr>))"

end

context
  fixes location_parser :: "'l xmlt"
    and trans_parser :: "'tr xmlt"
    and atom_parser :: "('f,'v,'t) exp formula xmlt"
    and hint_parser :: "'h :: default xmlt"
begin

type_synonym art_node_id = string

definition art_node_id_parser :: "tag \<Rightarrow> art_node_id xmlt" where
  "art_node_id_parser = Xmlt.text"

abbreviation art where "art \<equiv> art_parser art_node_id_parser location_parser trans_parser atom_parser hint_parser"

definition "invariant_proof_parser \<equiv>
  Xmlt.options [
    (''impact'', Xmlt.change art (\<lambda> prf i. Impact i prf))
  ]"

definition invariant_parser :: "('l \<times> ('f,'v,'t) exp formula) xmlt"
  where "invariant_parser \<equiv>
    Xmlt.pair ''invariant''
      (Xmlt.singleton ''location'' location_parser id)
      (Xmlt.singleton ''formula'' (formula_parser atom_parser) id) Pair"

definition invariants_parser :: "('l \<times> ('f,'v,'t) exp formula)list xmlt"
  where "invariants_parser = Xmlt.many ''invariants'' invariant_parser id"

definition safety_proof_parser :: "('f,'v,'t,'l,string,'tr,'h) safety_proof xmlt" where
 "safety_proof_parser \<equiv>
  Xmlt.pair ''safetyViaInvariants'' invariants_parser invariant_proof_parser
  (\<lambda> I p. safety_proof.Invariant_Assertion (p I) safety_proof.Trivial)"

end


fun cut_points_to_transitions :: "_ \<Rightarrow> _ \<Rightarrow> ('f,'v,'t,'l sharp,'tr sharp) transitions_impl" where
  "cut_points_to_transitions ts [] = ts"
| "cut_points_to_transitions ts ((l,tr,\<phi>)#cps) =
   cut_points_to_transitions ((Flat tr, Transition (Flat l) (Sharp l) \<phi>)#ts) cps"

definition xml_ignore :: "'a \<Rightarrow> 'a xmlt"
where
  "xml_ignore x xml \<equiv> return x"

context
  fixes location_id_parser :: "tag \<Rightarrow> 'l xmlt"
    and trans_id_parser :: "tag \<Rightarrow> 'tr xmlt"
    and exp_parser :: "('f,'v,'t) exp xmlt"
    and atom_parser :: "('f,'v,'t) exp formula xmlt"
    and tatom_parser :: "('f,'v trans_var,'t) exp formula xmlt"
    and variable_parser :: "'v xmlt"
    and hint_parser :: "'h :: default xmlt"
    and type_parser :: "'t xmlt"
    and dom_type :: "'t"
begin

definition "location_parser \<equiv> location_id_parser ''locationId''"

definition sharp_location_parser :: "'l sharp xmlt" where
  "sharp_location_parser =
    xml_if ''locationDuplicate''
      (Xmlt.change (location_id_parser ''locationDuplicate'') Sharp)
      (Xmlt.change (location_id_parser ''locationId'') Flat)"

abbreviation "tformula \<equiv> formula_parser tatom_parser"

context
  fixes trans_parser :: "'trr xmlt"
begin

abbreviation "lts \<equiv> lts_parser sharp_location_parser trans_parser tatom_parser"

abbreviation "transition \<equiv> transition_parser sharp_location_parser trans_parser tatom_parser"

abbreviation "invariant_proof \<equiv> invariant_proof_parser sharp_location_parser trans_parser atom_parser hint_parser"

abbreviation "invariants \<equiv> invariants_parser sharp_location_parser atom_parser"

partial_function(sum_bot) cooperation_proof_parser where
 [code]: "cooperation_proof_parser xml =
  Xmlt.options [
    (''trivial'', Xmlt.leaf ''trivial'' cooperation_proof.Trivial),
    (''newInvariants'',
      Xmlt.triple ''newInvariants''
        invariants
        invariant_proof
        cooperation_proof_parser
        (\<lambda> i p cp. Invariants_Update (p i) cp)
    ),
    (''transitionRemoval'', 
      Xmlt.tuple5 ''transitionRemoval''
       (Xmlt.many ''rankingFunctions''
          (Xmlt.pair ''rankingFunction''
            (Xmlt.singleton ''location'' sharp_location_parser id)
            (Xmlt.many1 ''expression'' exp_parser exp_parser Cons)
            (\<lambda>l es. Pair l es))
          id)
       (Xmlt.many1 ''bound'' exp_parser exp_parser Cons)
       (Xmlt.many ''remove'' trans_parser id)
       (Xmlt.many ''hints''
         (Xmlt.options [
           (''strictDecrease'', Xmlt.triple ''strictDecrease''
              trans_parser
              (Xmlt.many ''boundHints'' hint_parser id)
              (Xmlt.many ''decreaseHints'' hint_parser id)
              (\<lambda>x bnd dec. (x, default (*map (\<lambda>(y,z). [y,z]) (zip dec bnd)*)))
           ),
           (''weakDecrease'', Xmlt.pair ''weakDecrease''
              trans_parser
              (Xmlt.many ''decreaseHints'' hint_parser id)
              (\<lambda>x y. (x, default(*[y]*))))
         ]) (\<lambda>ps. map_of_default default ps))
       cooperation_proof_parser
       (\<lambda>rs bounds removed hinter inner. 
         Transition_Removal
           (Transition_removal_info
             (map_of_default [] rs)
             removed
             dom_type
             bounds
             hinter
           ) inner
       )
    ),
    (''locationAddition'', Xmlt.pair ''locationAddition''
        transition
        cooperation_proof_parser
        (\<lambda> (tr_id, tau) prof. Location_Addition (Location_Addition_Info (source tau) (target tau) tr_id tau) prof)),
   (''freshVariableAddition'', Xmlt.tuple4 ''freshVariableAddition''
       variable_parser
       type_parser
       (Xmlt.many ''additionalFormulas''
          (Xmlt.pair ''additionalFormula'' trans_parser tformula Pair) id)
       cooperation_proof_parser
        (\<lambda> x1 x2 x3 prof. Fresh_Variable_Addition (Fresh_Variable_Addition_Info x1 x2 x3) prof)),
   (''cutTransitionSplit'', Xmlt.many ''cutTransitionSplit''
       (Xmlt.pair ''cutTransitionsWithProof'' (Xmlt.many ''cutTransitions'' trans_parser id) cooperation_proof_parser Pair) Cut_Transition_Split),
   (''sccDecomposition'', Xmlt.many ''sccDecomposition''
       (Xmlt.pair ''sccWithProof'' (Xmlt.many ''scc'' sharp_location_parser id) cooperation_proof_parser Pair) Scc_Decomp)
  ] xml"

end

definition "trans_id \<equiv> trans_id_parser ''transitionId''"


definition "sharp_trans_id =
  xml_if ''transitionDuplicate''
    (Xmlt.change (trans_id_parser ''transitionDuplicate'') Sharp)
    (Xmlt.change (trans_id_parser ''transitionId'') Flat)"

definition "cutPoints_parser =
  Xmlt.many ''cutPoints''
    (Xmlt.triple ''cutPoint''
      (location_id_parser ''locationId'')
      (trans_id_parser ''skipId'')
      (Xmlt.singleton ''skipFormula'' tformula id)
      (\<lambda>x y z. (x,y,z))
    ) (\<lambda> z. (cut_points_to_transitions [] z))"

partial_function (sum_bot) termination_proof_parser where [code]: "termination_proof_parser xml =
  Xmlt.options [
    (''trivial'', Xmlt.leaf ''trivial'' termination_proof.Trivial),
    (''newInvariants'',
      Xmlt.triple ''newInvariants''
        (invariants_parser location_parser atom_parser)
        (invariant_proof_parser location_parser trans_id atom_parser hint_parser)
        termination_proof_parser
        (\<lambda> i p cp. Invariants_Update_LTS (p i) cp)
    ),
    (''switchToCooperationTermination'',
      Xmlt.pair ''switchToCooperationTermination''
        cutPoints_parser
        (cooperation_proof_parser sharp_trans_id)
        (\<lambda> cp p. termination_proof.Via_Cooperation [(cp,p)]))
    ] xml"

end


type_synonym location_id = string
type_synonym transition_id = string
type_synonym variable_id = string


definition variable_parser :: "variable_id xmlt" where
  "variable_parser = Xmlt.text ''variableId''"

definition trans_var_parser :: "variable_id trans_var xmlt" where
  "trans_var_parser =
    xml_enter_if ''post'' (xml_take variable_parser (xml_leave \<circ> Post))
      (Xmlt.change variable_parser Pre)"

context LIA begin

definition "constant_parser = Xmlt.int ''constant''"

definition sym_parser' :: "'a xmlt \<Rightarrow> 'a LIA_exp' xmlt" where
  "sym_parser' v =
    xml_enter_if ''constant'' (xml_take_int (xml_leave \<circ> (\<lambda> i. Fun (Const' i) [])))
    (Xmlt.change v (\<lambda> v. Var (v, IntegerT)))"

definition product_parser' :: "'a :: show xmlt \<Rightarrow> 'a LIA_exp' xmlt" where
  "product_parser' v xml \<equiv>
   case xml of
    XML tag [] [x,y] \<Rightarrow>
    if tag = ''product'' then
      do {
        l \<leftarrow> sym_parser' v x;
        r \<leftarrow> sym_parser' v y;
        case (l,r) of (Fun (Const' i) [],_) \<Rightarrow> return (Fun (Cmult' i) [r])
          | _ \<Rightarrow> error (show ''not accepting '' @ show x @ show '' * '' @ show y)
      }
    else
      sym_parser' v xml
  | _ \<Rightarrow>
      sym_parser' v xml"

definition "sum_parser' v =
  xml_if ''sum''
    (Xmlt.many ''sum'' (product_parser' v) (Fun Sum'))
    (product_parser' v)"

definition bexp_parser' :: "'a :: show xmlt \<Rightarrow> 'a LIA_exp' xmlt" where
  "bexp_parser' v = Xmlt.change (Xmlt.options [
    (''leq'', Xmlt.pair ''leq'' (sum_parser' v) (sum_parser' v) (\<lambda> l r. Fun Lesseq' [l, r])),
    (''less'', Xmlt.pair ''less'' (sum_parser' v) (sum_parser' v) (\<lambda> l r. Fun Less' [l,r])),
    (''eq'', Xmlt.pair ''eq'' (sum_parser' v) (sum_parser' v) (\<lambda> l r. Fun Equals' [l,r]))
  ]) id"

definition "exp_parser \<equiv> Xmlt.change (sum_parser' variable_parser) of_LIA_exp'"
definition "atom_parser \<equiv> Xmlt.change (bexp_parser' variable_parser) of_LIA_atom'"
definition "tatom_parser \<equiv> Xmlt.change (bexp_parser' trans_var_parser) of_LIA_atom'"
definition "atom_parser' = Xmlt.change (bexp_parser' variable_parser) Atom"
definition "tatom_parser' = Xmlt.change (bexp_parser' trans_var_parser) Atom"

definition hint_parser :: "lia_hints xmlt" where
  "hint_parser = Xmlt.singleton ''linearImplicationHint''
    (Xmlt.options [
    (''linearCombination'', Xmlt.many ''linearCombination'' constant_parser (\<lambda> xs. LIA_Hints (map nat xs))),
    (''simplex'', Xmlt.leaf ''simplex'' LIA_Simplex)
     ]) id"

definition type_parser :: "ty xmlt" where
  "type_parser = Xmlt.options [(''int'', Xmlt.leaf ''int'' IntegerT), (''bool'', Xmlt.leaf ''bool'' BoolT)]"


definition "lts_input_parser \<equiv> lts_parser (location_parser Xmlt.text) (trans_id Xmlt.text) tatom_parser' ''lts''"

definition "lts_safety_input_parser \<equiv> safety_input_parser (location_parser Xmlt.text) (trans_id Xmlt.text) tatom_parser'"

definition "lts_termination_proof_parser \<equiv>
  termination_proof_parser
    Xmlt.text Xmlt.text exp_parser atom_parser tatom_parser variable_parser
    hint_parser type_parser IntegerT"

definition "lts_safety_proof_parser \<equiv> safety_proof_parser (location_parser Xmlt.text) (trans_id Xmlt.text) atom_parser hint_parser"

end

end