(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2011-2015)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2013)
License: LGPL (see file COPYING.LESSER)
*)
theory Proof_Checker
imports
  Check_Termination
  Check_Nontermination
  Check_CRP
  Check_Conditional_CRP
  Check_Completion_Proof
  Check_Equational_Proof
  Check_Complexity
  Check_Quasi_Reductive
  Check_AC_Termination
  "../LTS/LIA_Instance" 
begin

datatype ('f, 'v) strategy =
  No_Strategy
| Innermost
| Innermost_Q "('f, 'v) term list"

primrec strategy_to_Q :: "('f, 'v) strategy \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term list" where
  "strategy_to_Q No_Strategy _ = []"
| "strategy_to_Q Innermost R = remdups (map fst R)"
| "strategy_to_Q (Innermost_Q Q) _ = Q"

fun rel_rules_of :: "('f,'v)rules option \<Rightarrow> ('f,'v)rules"
 where "rel_rules_of None = []"
     | "rel_rules_of (Some r) = r"

datatype ('f, 'v) input_problem =
  TRS bool "('f, 'v) strategy" "('f, 'v) rules" "('f, 'v) rules"
| OTRS "('f, 'v) rules"
| CSTRS "'f muL" "('f, 'v) rules"
| FPTRS "('f, 'v) forb_pattern list" "('f, 'v) rules"
| DP bool bool "('f, 'v) rules" "('f, 'v) rules" "('f, 'v) strategy" "('f, 'v) rules" "('f, 'v) rules"
| EQS_TRS "('f, 'v) equation list" "('f, 'v) rules"
| EQS_EQ "('f, 'v) equation list" "('f, 'v) term \<times> ('f, 'v) term"
| Complexity_TRS "('f, 'v) strategy" "('f, 'v) rules" "('f, 'v) rules" "('f, 'v) complexity_measure" complexity_class
| CTRS "('f, 'v) crules"
| TA_Problem "(string,'f)tree_automaton" "('f,'v)rules"
| AC_Problem "('f,'v) rules" "'f list" "'f list"
| LTS_Termination_Problem "(LIA.sig', 'v, LIA.ty, string, string) lts_impl"
| LTS_Safety_Problem "(LIA.sig', 'v, LIA.ty, string, string) lts_impl" "string list" 
| Unknown unknown_info

datatype ('f, 'l, 'v) cert_problem = 
  TRS_Termination_Proof bool "(('f, 'l) lab, 'v) strategy"
            "(('f, 'l) lab, 'v) rules"
            "(('f, 'l) lab, 'v) rules option"
            "('f, 'l, 'v) trs_termination_proof"
| Complexity_Proof "(('f, 'l) lab, 'v) strategy"
            "(('f, 'l) lab, 'v) rules"
            "(('f, 'l) lab, 'v) rules option"
            "(('f,'l)lab,'v)complexity_measure" complexity_class
            "('f,'l,'v) complexity_proof"
| DP_Termination_Proof bool bool "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) strategy"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) rules"
  "('f, 'l, 'v) dp_termination_proof"
| DP_Nontermination_Proof bool bool "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) strategy"
  "(('f, 'l) lab, 'v) rules"
  "('f, 'l, 'v) dp_nontermination_proof"
| TRS_Nontermination_Proof bool "(('f, 'l) lab, 'v) strategy"
               "(('f, 'l) lab, 'v) rules"
               "('f, 'l, 'v) trs_nontermination_proof"
| Outermost_Termination_Proof 
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fptrs_termination_proof"
| Outermost_Nontermination_Proof 
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fp_nontermination_proof"
| CS_Termination_Proof "('f,'l)lab muL"
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fptrs_termination_proof"
| CS_Nontermination_Proof "('f,'l)lab muL"
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fp_nontermination_proof"
| FP_Termination_Proof "(('f, 'l) lab, 'v) forb_pattern list"
               "(('f, 'l) lab, 'v) rules"
               "('f, 'l, 'v) fptrs_termination_proof"
| FP_Nontermination_Proof "(('f, 'l) lab, 'v) forb_pattern list"
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fp_nontermination_proof"
| Relative_TRS_Nontermination_Proof bool "(('f, 'l) lab, 'v) strategy"
                   "(('f, 'l) lab, 'v) rules"
                   "(('f, 'l) lab, 'v) rules"
                   "('f, 'l, 'v) reltrs_nontermination_proof"
| TRS_Confluence_Proof bool "(('f, 'l) lab, 'v) rules"
                 "('f, 'l, 'v) cr_proof"
| TRS_Non_Confluence_Proof bool "(('f, 'l) lab, 'v) rules"
                 "('f, 'l, 'v, 'v) ncr_proof"
| Completion_Proof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) rules"
                   "('f, 'l, 'v) completion_proof"
| Equational_Proof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) term \<times> (('f, 'l) lab, 'v) term"
                   "('f, 'l, 'v) equational_proof"
| Equational_Disproof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) term \<times> (('f, 'l) lab, 'v) term"
                   "('f, 'l, 'v) equational_disproof"
| Quasi_Reductive_Proof "(('f, 'l) lab, 'v) crules"
                   "('f, 'l, 'v) quasi_reductive_proof"
| Conditional_CR_Proof "(('f, 'l) lab, 'v) crules"
                   "('f, 'l, 'v) conditional_cr_proof"
| Conditional_Non_CR_Proof "(('f, 'l) lab, 'v) crules"
                   "('f, 'l, 'v, 'v) conditional_ncr_proof"
| Tree_Automata_Closed_Proof "(string,('f,'l)lab)tree_automaton" "(('f,'l)lab,'v)rules" "string ta_relation"
| AC_Termination_Proof "(('f,'l)lab, 'v) rules"
                   "('f,'l)lab list"
                   "('f,'l)lab list"
                   "('f,'l,'v) ac_termination_proof"
| LTS_Termination_Proof "(LIA.sig', 'v, LIA.ty, string, string) lts_impl"
             "(LIA.sig, 'v, LIA.ty, string, string, lia_hints) termination_proof"
| LTS_Safety_Proof
    "(LIA.sig', 'v, LIA.ty, string, string) lts_impl" "string list"
    "(LIA.sig, 'v, LIA.ty, string, string, string, lia_hints) safety_proof"
| Unknown_Proof unknown_info "('f,'l,'v) unknown_proof"
| Unknown_Disproof unknown_info "('f,'l,'v) neg_unknown_proof"

hide_const (open) eq (* eq from HOL-Algebra congruence is here used as variable name for an equation *)

primrec input :: "('f, 'l, 'v) cert_problem \<Rightarrow> (('f, 'l) lab, 'v) input_problem" where
  "input (TRS_Termination_Proof nfs Q R S_o _) = TRS nfs Q R (rel_rules_of S_o)"
| "input (Complexity_Proof Q R S_o cm cc _) = Complexity_TRS Q R (rel_rules_of S_o) cm cc"
| "input (TRS_Nontermination_Proof nfs Q R _) = TRS nfs Q R []"
| "input (Outermost_Nontermination_Proof R _) = OTRS R"
| "input (Outermost_Termination_Proof R _) = OTRS R"
| "input (CS_Nontermination_Proof \<mu> R _) = CSTRS \<mu> R"
| "input (CS_Termination_Proof \<mu> R _) = CSTRS \<mu> R"
| "input (FP_Termination_Proof P R _) = FPTRS P R"
| "input (FP_Nontermination_Proof P R _) = FPTRS P R"
| "input (Relative_TRS_Nontermination_Proof nfs Q R S _) = TRS nfs Q R S"
| "input (DP_Termination_Proof nfs m P Pw Q R Rw _) = DP nfs m P Pw Q R Rw"
| "input (DP_Nontermination_Proof nfs m P Q R _) = DP nfs m P [] Q [] R"
| "input (TRS_Confluence_Proof nfs R _) = TRS nfs No_Strategy R []"
| "input (TRS_Non_Confluence_Proof nfs R _) = TRS nfs No_Strategy R []"
| "input (Completion_Proof E R _) = EQS_TRS E R"
| "input (Equational_Proof E eq _) = EQS_EQ E eq"
| "input (Equational_Disproof E eq _) = EQS_EQ E eq"
| "input (Quasi_Reductive_Proof ctrs _) = CTRS ctrs"
| "input (Conditional_CR_Proof ctrs _) = CTRS ctrs"
| "input (Conditional_Non_CR_Proof ctrs _) = CTRS ctrs"
| "input (Tree_Automata_Closed_Proof ta R _) = TA_Problem ta R"
| "input (AC_Termination_Proof r a c _) = AC_Problem r a c"
| "input (LTS_Termination_Proof r _) = LTS_Termination_Problem r"
| "input (LTS_Safety_Proof r e _) = LTS_Safety_Problem r e" 
| "input (Unknown_Proof up _) = Unknown up"
| "input (Unknown_Disproof up _) = Unknown up"

primrec desired_property :: "('f, 'l, 'v) cert_problem \<Rightarrow> bool" where
  "desired_property (TRS_Termination_Proof nfs Q R S_o _ ) = SN_qrel (nfs, set (strategy_to_Q Q R), set R, set (rel_rules_of S_o))"
| "desired_property (Complexity_Proof Q R S_o cm cc _ ) = deriv_bound_measure_class (rel_qrstep (True, set (strategy_to_Q Q (R @ rel_rules_of S_o)), set R, set (rel_rules_of S_o))) cm cc"
| "desired_property (TRS_Nontermination_Proof nfs Q R _) = (\<not> SN (qrstep nfs (set (strategy_to_Q Q R)) (set R)))"
| "desired_property (Outermost_Nontermination_Proof R _) = (\<not> SN (ostep (lhss (set R)) (set R)))"
| "desired_property (Outermost_Termination_Proof R _) = (SN (ostep (lhss (set R)) (set R)))"
| "desired_property (FP_Nontermination_Proof P R _) = (\<not> SN (fpstep (set P) (set R)))"
| "desired_property (FP_Termination_Proof P R _) = (SN (fpstep (set P) (set R)))"
| "desired_property (CS_Nontermination_Proof \<mu> R _) = (\<not> SN (csstep (mu_of \<mu>) (set R)))"
| "desired_property (CS_Termination_Proof \<mu> R _) = (SN (csstep (mu_of \<mu>) (set R)))"
| "desired_property (Relative_TRS_Nontermination_Proof nfs Q R S _) = (\<not> SN_qrel (nfs, set (strategy_to_Q Q R), set R, set S))"
| "desired_property (DP_Termination_Proof nfs m P Pw Q R Rw _) = finite_dpp (nfs, m, set P, set Pw, set (strategy_to_Q Q (R @ Rw)), set R, set Rw)"
| "desired_property (DP_Nontermination_Proof nfs m P Q R _) = infinite_dpp (nfs, set P, set (strategy_to_Q Q R), set R)" (*ignores nfs and m flags!*)
| "desired_property (TRS_Confluence_Proof nfs R _) = CR (rstep (set R))"
| "desired_property (TRS_Non_Confluence_Proof nfs R _) = (\<not> CR (rstep (set R)))"
| "desired_property (Completion_Proof E R _) = completed_rewrite_system (set E) (set R)"
| "desired_property (Equational_Proof E eq _) = (eq \<in> (estep (set E))^*)"
| "desired_property (Equational_Disproof E eq _) = (eq \<notin> (estep (set E))^*)"
| "desired_property (Quasi_Reductive_Proof ctrs _) = (quasi_reductive (set ctrs))"
| "desired_property (Conditional_CR_Proof ctrs _) = (CR (cstep (set ctrs)))"
| "desired_property (Conditional_Non_CR_Proof ctrs _) = (\<not> (CR (cstep (set ctrs))))"
| "desired_property (Tree_Automata_Closed_Proof ta R _) = ((rstep (set R))^* `` ta_lang (ta_of_ta ta) \<subseteq> ta_lang (ta_of_ta ta))"
| "desired_property (AC_Termination_Proof r a c _) = (SN (relation_ac_tp (set r, set a, set c)))"
| "desired_property (LTS_Termination_Proof r _) = (SN_on (prelogic.transition LIA.\<Lambda> (lts_of' r)) (prelogic.initial_states LIA.\<Lambda> (lts_of' r)))"
| "desired_property (LTS_Safety_Proof r e _) = (prelogic.lts_safe LIA.\<Lambda> (lts_of' r) (set e))"
| "desired_property (Unknown_Proof up _) = (unknown_satisfied up)"
| "desired_property (Unknown_Disproof up _) = (unknown_satisfied up)"

fun xml_tag :: "('f, 'l, 'v) cert_problem \<Rightarrow> string" where
  "xml_tag (TRS_Termination_Proof _ _ _ (Some _) _ ) = ''<relativeTerminationProof>''"
| "xml_tag (TRS_Termination_Proof _ _ _ None _ ) = ''<trsTerminationProof>''"
| "xml_tag (Complexity_Proof _ _ _ _ _ _ ) = ''<complexityProof>''"
| "xml_tag (TRS_Nontermination_Proof _ _ _ _) = ''<trsNonterminationProof>''"
| "xml_tag (Outermost_Nontermination_Proof _ _) = ''<trsNonterminationProof>''"
| "xml_tag (Outermost_Termination_Proof _ _) = ''<trsTerminationProof>''"
| "xml_tag (FP_Nontermination_Proof _ _ _) = ''<trsNonterminationProof>''"
| "xml_tag (FP_Termination_Proof _ _ _) = ''<trsTerminationProof>''"
| "xml_tag (CS_Nontermination_Proof _ _ _) = ''<trsNonterminationProof>''"
| "xml_tag (CS_Termination_Proof _ _ _) = ''<trsTerminationProof>''"
| "xml_tag (Relative_TRS_Nontermination_Proof _ _ _ _ _) = ''<relativeNonterminationProof>''"
| "xml_tag (DP_Termination_Proof _ _ _ _ _ _ _ _) = ''<dpProof>''"
| "xml_tag (DP_Nontermination_Proof _ _ _ _ _ _) = ''<dpNonterminationProof>''"
| "xml_tag (TRS_Confluence_Proof _ _ _) = ''<crProof>''"
| "xml_tag (TRS_Non_Confluence_Proof _ _ _) = ''<crDisproof>''"
| "xml_tag (Completion_Proof _ _ _) = ''<completionProof>''"
| "xml_tag (Equational_Proof _ _ _) = ''<equationalProof>''"
| "xml_tag (Equational_Disproof _ _ _) = ''<equationalDisproof>''"
| "xml_tag (Quasi_Reductive_Proof _ _) = ''<quasiReductiveProof>''"
| "xml_tag (Conditional_CR_Proof _ _) = ''<conditionalCrProof>''"
| "xml_tag (Conditional_Non_CR_Proof _ _) = ''<conditionalCrDisproof>''"
| "xml_tag (Tree_Automata_Closed_Proof _ _ _) = ''<treeAutomatonClosedProof>''"
| "xml_tag (AC_Termination_Proof _ _ _ _) = ''<acTerminationProof>''"
| "xml_tag (LTS_Termination_Proof _ _) = ''<ltsTerminationProof>''"
| "xml_tag (LTS_Safety_Proof _ _ _) = ''<ltsSafetyProof>''"
| "xml_tag (Unknown_Proof _ _) = ''<unknownInputProof>''"
| "xml_tag (Unknown_Disproof _ _) = ''<unknownInputProof>''"

fun cert_assms :: "bool \<Rightarrow> ('f, 'l, 'v) cert_problem \<Rightarrow> (('f, 'l, 'v) assm) list" where
  "cert_assms a (TRS_Termination_Proof _ _ _ _ p) = sn_assms a p"
| "cert_assms a (Outermost_Termination_Proof _ p) = sn_fp_assms a p"
| "cert_assms a (Outermost_Nontermination_Proof _ p) = neg_fp_sn_assms a p"
| "cert_assms a (FP_Termination_Proof _ _ p) = sn_fp_assms a p"
| "cert_assms a (FP_Nontermination_Proof _ _ p) = neg_fp_sn_assms a p"
| "cert_assms a (CS_Termination_Proof _ _ p) = sn_fp_assms a p"
| "cert_assms a (CS_Nontermination_Proof _ _ p) = neg_fp_sn_assms a p"
| "cert_assms a (TRS_Confluence_Proof _ _ p) = cr_assms a p"
| "cert_assms a (TRS_Non_Confluence_Proof _ _ p) = ncr_assms a p"
| "cert_assms a (Completion_Proof _ _ p) = completion_assms a p"
| "cert_assms a (DP_Termination_Proof _ _ _ _ _ _ _ p) = dp_assms a p"
| "cert_assms a (Equational_Proof _ _ p) = equational_assms a p"
| "cert_assms a (Equational_Disproof _ _ p) = equational_dis_assms a p"
| "cert_assms a (Quasi_Reductive_Proof _ p) = quasi_reductive_assms a p"
| "cert_assms a (Conditional_CR_Proof _ p) = conditional_cr_assms a p"
| "cert_assms a (Conditional_Non_CR_Proof _ p) = conditional_ncr_assms a p"
| "cert_assms a (Unknown_Proof _ p) = unknown_assms a p"
| "cert_assms a (Unknown_Disproof _ p) = neg_unknown_assms a p"
| "cert_assms a (Relative_TRS_Nontermination_Proof _ _ _ _ p) = neg_rel_sn_assms a p"
| "cert_assms a (TRS_Nontermination_Proof _ _ _ p) = neg_sn_assms a p"
| "cert_assms a (DP_Nontermination_Proof _ _ _ _ _ p) = neg_dp_assms a p"
| "cert_assms a (Complexity_Proof _ _ _ _ _ p ) =  complexity_assms a p"
| "cert_assms a (LTS_Termination_Proof _ _) = []"
| "cert_assms a (LTS_Safety_Proof _ _ _) = []"
| "cert_assms a (Tree_Automata_Closed_Proof _ _ _) =  []"
| "cert_assms a (AC_Termination_Proof _ _ _ _) =  []"

lemma cert_assms_False[simp]: "cert_assms False p = []" by (cases p) simp_all

datatype cert_result = Certified string | Unsupported string | Error string

context fixes
  I :: "('tp, ('f::{show,key,countable}, label_type) lab, string) tp_ops" and
  J :: "('dpp, ('f, label_type) lab, string) dpp_ops" and
  K :: "('ac_tp, ('f, label_type) lab, string) ac_tp_ops" and
  L :: "('ac_dpp, ('f, label_type) lab, string) ac_dpp_ops"
begin

fun
  certify_cert_problem ::
    "bool \<Rightarrow> ('f, label_type, string) cert_problem \<Rightarrow> cert_result"
where
  "certify_cert_problem a (TRS_Termination_Proof nfs Q R None prf) =
    (case check_trs_termination_proof I J a (shows_string ''1'') (tp_ops.mk I nfs (strategy_to_Q Q R) R []) prf of
      Inl err \<Rightarrow> Error ((''termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Termination_Proof nfs Q R None prf)))"
| "certify_cert_problem a (TRS_Termination_Proof nfs Q R (Some S) prf) = 
    (case check_trs_termination_proof I J a (shows_string ''1'') (tp_ops.mk I nfs (strategy_to_Q Q R) R S) prf of
      Inl err \<Rightarrow> Error ((''relative termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Termination_Proof nfs Q R (Some S) prf)))"
| "certify_cert_problem a (Complexity_Proof Q R S_o cm cc prf) = 
    (case check_complexity_proof I a (shows_string ''1'') (tp_ops.mk I True (strategy_to_Q Q (R @ rel_rules_of S_o)) R (rel_rules_of S_o), cm, cc) prf of
      Inl err \<Rightarrow> Error ((''complexity proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Complexity_Proof Q R S_o cm cc prf)))"
| "certify_cert_problem a (TRS_Nontermination_Proof nfs Q R prf) =
    (case check_trs_nontermination_proof I J a (shows_string ''1'') (tp_ops.mk I nfs (strategy_to_Q Q R) R []) prf of
      Inl err \<Rightarrow> Error ((''nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Nontermination_Proof nfs Q R prf)))"
| "certify_cert_problem a (Outermost_Nontermination_Proof R prf) =
    (case check_fp_nontermination_proof I J a (shows_string ''1'') (o_to_fp_impl (map fst R),R) prf of
      Inl err \<Rightarrow> Error ((''outermost nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Outermost_Nontermination_Proof R prf)))"
| "certify_cert_problem a (Outermost_Termination_Proof R prf) =
    (case check_fptrs_termination_proof I J a (shows_string ''1'') (o_to_fp_impl (map fst R), R) prf of
      Inl err \<Rightarrow> Error ((''outermost termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Outermost_Termination_Proof R prf)))"
| "certify_cert_problem a (CS_Nontermination_Proof \<mu> R prf) =
    (case check_fp_nontermination_proof  I J a (shows_string ''1'') (mu_to_fp_impl \<mu>,R) prf of
      Inl err \<Rightarrow> Error ((''outermost nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (CS_Nontermination_Proof \<mu> R prf)))"
| "certify_cert_problem a (CS_Termination_Proof \<mu> R prf) =
    (case check_fptrs_termination_proof I J a (shows_string ''1'') (mu_to_fp_impl \<mu>, R) prf of
      Inl err \<Rightarrow> Error ((''outermost termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (CS_Termination_Proof \<mu> R prf)))"
| "certify_cert_problem a (FP_Termination_Proof P R prf) =
    (case check_fptrs_termination_proof I J a (shows_string ''1'') (P, R) prf of
      Inl err \<Rightarrow> Error ((''forbidden pattern termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (FP_Termination_Proof P R prf)))"
| "certify_cert_problem a (FP_Nontermination_Proof P R prf) =
    (case check_fp_nontermination_proof  I J a (shows_string ''1'') (P,R) prf of
      Inl err \<Rightarrow> Error ((''forbidden pattern nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (FP_Nontermination_Proof P R prf)))"
| "certify_cert_problem a (Relative_TRS_Nontermination_Proof nfs Q R S prf) =
    (case check_reltrs_nontermination_proof I J a (shows_string ''1'') (tp_ops.mk I nfs (strategy_to_Q Q R) R S) prf of
      Inl err \<Rightarrow>
        Error ((''relative nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Relative_TRS_Nontermination_Proof nfs Q R S prf)))"
| "certify_cert_problem a (DP_Termination_Proof nfs m P Pw Q R Rw prf) =
    (case check_dp_termination_proof I J a (shows_string ''1'') (dpp_ops.mk J nfs m P Pw (strategy_to_Q Q (R @ Rw)) R Rw) prf of
      Inl err \<Rightarrow> Error ((''finiteness proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (DP_Termination_Proof nfs m P Pw Q R Rw prf)))"
| "certify_cert_problem a (DP_Nontermination_Proof nfs m P Q R prf) =
    (case check_dp_nontermination_proof I J a (shows_string ''1'') (dpp_ops.mk J nfs m P [] (strategy_to_Q Q R) [] R) prf of
      Inl err \<Rightarrow> Error ((''infiniteness proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (DP_Nontermination_Proof nfs m P Q R prf)))"
| "certify_cert_problem a (TRS_Confluence_Proof nfs R prf) =
    (case check_cr_proof a (shows_string ''1'') I J R prf of
      Inl err \<Rightarrow> Error ((''confluence proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Confluence_Proof nfs R prf)))"
| "certify_cert_problem a (TRS_Non_Confluence_Proof nfs R prf) =
    (case check_ncr_proof a (shows_string ''1'') I J R prf of
      Inl err \<Rightarrow> Error ((''confluence disproof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Non_Confluence_Proof nfs R prf)))"
| "certify_cert_problem a (Completion_Proof E R prf) =
    (case check_completion_proof a (shows_string ''1'') I J E R prf of
      Inl err \<Rightarrow> Error ((''completion proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Completion_Proof E R prf)))"
| "certify_cert_problem a (Equational_Proof E eq prf) =
    (case check_equational_proof a (shows_string ''1'') I J E eq prf of
      Inl err \<Rightarrow> Error ((''equational proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Equational_Proof E eq prf)))"
| "certify_cert_problem a (Equational_Disproof E eq prf) =
    (case check_equational_disproof a (shows_string ''1'') I J E eq prf of
      Inl err \<Rightarrow> Error ((''equational disproof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Equational_Disproof E eq prf)))"
| "certify_cert_problem a (Quasi_Reductive_Proof ctrs prf) =
    (case check_quasi_reductive_proof a (shows_string ''1'') I J ctrs prf of
      Inl err \<Rightarrow> Error ((''quasi reductive proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Quasi_Reductive_Proof ctrs prf)))"
| "certify_cert_problem a (Conditional_CR_Proof ctrs prf) =
    (case check_conditional_cr_proof a (shows_string ''1'') I J ctrs prf of
      Inl err \<Rightarrow> Error ((''conditional confluence proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Conditional_CR_Proof ctrs prf)))"
| "certify_cert_problem a (Conditional_Non_CR_Proof ctrs prf) =
    (case check_conditional_ncr_proof a (shows_string ''1'') I J ctrs prf of
      Inl err \<Rightarrow> Error ((''conditional confluence disproof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Conditional_Non_CR_Proof ctrs prf)))"
| "certify_cert_problem a (Tree_Automata_Closed_Proof ta R prf) =
    (case tree_aut_trs_closed ta prf R of
      Inl err \<Rightarrow> Error ((''tree automaton closed proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Tree_Automata_Closed_Proof ta R prf)))"
| "certify_cert_problem a (AC_Termination_Proof R A C prf) =
    (case check_ac_termination_proof L K (shows_string ''1'') (ac_tp_ops.mk K R A C) prf of
      Inl err \<Rightarrow> Error ((''AC termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (AC_Termination_Proof R A C prf)))"
| "certify_cert_problem a (LTS_Termination_Proof r prf) =
    (case LIA_logic.check_termination (of_LIA'_lts r) prf of
      Inl err \<Rightarrow> Error ((''LTS termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (LTS_Termination_Proof r prf :: ('f, label_type, string) cert_problem)))"
| "certify_cert_problem a (LTS_Safety_Proof r e prf) =
    (case LIA_logic.check_safety (of_LIA'_lts r) e prf of
      Inl err \<Rightarrow> Error ((''LTS safety proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (LTS_Safety_Proof r e prf :: ('f, label_type, string) cert_problem)))"
| "certify_cert_problem a (Unknown_Proof up prf) =
    (case check_unknown_proof I J a (shows_string ''1'') up prf of
      Inl err \<Rightarrow> Error ((''unknown proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Unknown_Proof up prf)))"
| "certify_cert_problem a (Unknown_Disproof up prf) =
    (case check_unknown_disproof I J a (shows_string ''1'') up prf of
      Inl err \<Rightarrow> Error ((''unknown proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Unknown_Disproof up prf)))"

context
  assumes I: "tp_spec I" and J: "dpp_spec J" and K: "ac_tp_spec K" and L: "ac_dpp_spec L"
begin

lemma certify_cert_problem_with_assms_sound:
  assumes answer: "certify_cert_problem a cp = Certified info"
  shows "((\<forall>p\<in>set (cert_assms a cp). holds p) \<longrightarrow> desired_property cp) \<and> xml_tag cp = info"
proof -
  let ?i = "shows_string ''1''"
  interpret tp_spec I by (rule I)  
  show ?thesis
  proof (cases cp)
    case (TRS_Termination_Proof nfs Q R sOpt prof)
    let ?tp = "tp_ops.mk I nfs (strategy_to_Q Q R) R"
    show ?thesis
    proof (cases sOpt)
      case None
      from answer have ok: "isOK (check_trs_termination_proof I J a ?i (?tp []) prof) \<and> xml_tag cp = info" 
        unfolding TRS_Termination_Proof None
        by (cases "check_trs_termination_proof I J a ?i (?tp []) prof") auto
      from check_trs_termination_proof_with_assms[OF I J ok[THEN conjunct1]]
       and ok show ?thesis unfolding TRS_Termination_Proof None by simp
    next
      case (Some S)
      from answer have ok: "isOK (check_trs_termination_proof I J a ?i (?tp S) prof) \<and> xml_tag cp = info"
        unfolding TRS_Termination_Proof Some
        by (cases "check_trs_termination_proof I J a ?i (?tp S) prof", auto)
      from check_trs_termination_proof_with_assms[OF I J ok[THEN conjunct1]] 
        and ok show ?thesis unfolding TRS_Termination_Proof Some by simp
    qed
  next
    case (Complexity_Proof Q R S_o cm cc prof)
    let ?Q = "strategy_to_Q Q (R @ rel_rules_of S_o)"
    let ?S = "rel_rules_of S_o"
    let ?tp = "tp_ops.mk I True ?Q R ?S"
    from answer have ok: "isOK (check_complexity_proof I a ?i (?tp,cm,cc) prof) \<and> xml_tag cp = info" 
      unfolding Complexity_Proof 
      by (cases "check_complexity_proof I a ?i (?tp,cm,cc) prof") auto
    from check_complexity_proof_with_assms[OF I infinite_lab ok[THEN conjunct1]]
      and ok show ?thesis unfolding Complexity_Proof by simp
  next
    case (TRS_Nontermination_Proof nfs Q R "prf")
    let ?tp = "tp_ops.mk I nfs (strategy_to_Q Q R) R []"
    from answer have ok: "isOK (check_trs_nontermination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding TRS_Nontermination_Proof
      by (cases "check_trs_nontermination_proof I J a ?i ?tp prf") auto
    from check_trs_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps TRS_Nontermination_Proof by simp
  next
    case (Outermost_Termination_Proof R "prf")
    have id: "lhss (set R) = set (map fst R)" by auto
    let ?tp = "(o_to_fp_impl (map fst R),R)"
    from answer have ok: "isOK (check_fptrs_termination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding Outermost_Termination_Proof
      by (cases "check_fptrs_termination_proof I J a ?i ?tp prf") auto
    from check_fptrs_termination_proof_with_assms[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding Outermost_Termination_Proof 
      by (simp add: SN_fpstep_def, unfold id ostep_iff_fpstep_impl, simp)
  next
    case (Outermost_Nontermination_Proof R "prf")
    have id: "lhss (set R) = set (map fst R)" by auto
    let ?p = "o_to_fp_impl (map fst R)"
    from answer have ok: "isOK (check_fp_nontermination_proof I J a ?i (?p,R) prf) \<and> xml_tag cp = info"
      unfolding Outermost_Nontermination_Proof
      by (cases "check_fp_nontermination_proof I J a ?i (?p,R) prf") auto
    from check_fp_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps Outermost_Nontermination_Proof 
      by (simp add: SN_fpstep_def, unfold id ostep_iff_fpstep_impl, simp)
  next
    case (CS_Termination_Proof \<mu> R "prf")
    let ?tp = "(mu_to_fp_impl \<mu>,R)"
    from answer have ok: "isOK (check_fptrs_termination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding CS_Termination_Proof
      by (cases "check_fptrs_termination_proof I J a ?i ?tp prf") auto
    from check_fptrs_termination_proof_with_assms[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding CS_Termination_Proof by (simp add: SN_fpstep_def csstep_iff_fpstep_impl)
  next
    case (CS_Nontermination_Proof \<mu> R "prf")
    let ?p = "mu_to_fp_impl \<mu>"
    from answer have ok: "isOK (check_fp_nontermination_proof I J a ?i (?p,R) prf) \<and> xml_tag cp = info"
      unfolding CS_Nontermination_Proof
      by (cases "check_fp_nontermination_proof I J a ?i (?p,R) prf") auto
    from check_fp_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps CS_Nontermination_Proof by (simp add: SN_fpstep_def csstep_iff_fpstep_impl)
  next
    case (FP_Termination_Proof P R "prf")
    let ?tp = "(P,R)"
    from answer have ok: "isOK (check_fptrs_termination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding FP_Termination_Proof
      by (cases "check_fptrs_termination_proof I J a ?i ?tp prf") auto
    from check_fptrs_termination_proof_with_assms[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding FP_Termination_Proof by (simp add: SN_fpstep_def)
  next
    case (FP_Nontermination_Proof P R "prf")
    from answer have ok: "isOK (check_fp_nontermination_proof I J a ?i (P,R) prf) \<and> xml_tag cp = info"
      unfolding FP_Nontermination_Proof
      by (cases "check_fp_nontermination_proof I J a ?i (P,R) prf") auto
    from check_fp_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps FP_Nontermination_Proof by (simp add: SN_fpstep_def)
  next
    case (Relative_TRS_Nontermination_Proof nfs Q R S "prf")
    let ?tp = "tp_ops.mk I nfs (strategy_to_Q Q R) R S"
    from answer have ok: "isOK (check_reltrs_nontermination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding Relative_TRS_Nontermination_Proof
      by (cases "check_reltrs_nontermination_proof I J a ?i ?tp prf") auto
    from check_reltrs_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps Relative_TRS_Nontermination_Proof by simp
  next
    case (DP_Termination_Proof nfs m P Pw Q R Rw "prf")
    let ?dp = "dpp_ops.mk J nfs m P Pw (strategy_to_Q Q (R @ Rw)) R Rw"
    from answer have ok: "isOK (check_dp_termination_proof I J a ?i ?dp prf) \<and> xml_tag cp = info"
      unfolding DP_Termination_Proof
      by (cases "check_dp_termination_proof I J a ?i ?dp prf", auto)
    interpret J: dpp_spec J by (rule J)
    show ?thesis
    proof (intro conjI impI)
      assume ass: "\<forall> p \<in> set (cert_assms a cp). holds p"
      show "desired_property cp"
        unfolding DP_Termination_Proof desired_property.simps
      proof (rule check_dp_termination_proof_with_assms[OF I J ok[THEN conjunct1], unfolded 
            J.dpp_spec_sound J.mk_simps cert_assms.simps])
        from ass
        show "\<forall>p \<in> set (dp_assms a prf). holds p"
          unfolding DP_Termination_Proof by simp
      qed
    qed (insert ok, auto)
  next
    case (DP_Nontermination_Proof nfs m P Q R "prf")
    let ?dp = "dpp_ops.mk J nfs m P [] (strategy_to_Q Q R) [] R"
    from answer have ok: "isOK (check_dp_nontermination_proof I J a ?i ?dp prf) \<and> xml_tag cp = info"
      unfolding DP_Nontermination_Proof
      by (cases "check_dp_nontermination_proof I J a ?i ?dp prf", auto)
    interpret J: dpp_spec J by (rule J)
    show ?thesis
    proof (intro conjI impI)
      assume ass: "\<forall> p \<in> set (cert_assms a cp). holds p"
      show "desired_property cp"
        unfolding DP_Nontermination_Proof desired_property.simps
        using check_dp_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1], unfolded 
            J.dpp_spec_sound J.mk_simps]
        using DP_Nontermination_Proof ass
        by auto
    qed (insert ok, auto)
  next
    case (TRS_Confluence_Proof nfs R "prf")
    from answer have ok: "isOK (check_cr_proof a ?i I J R prf) \<and> xml_tag cp = info"
      unfolding TRS_Confluence_Proof
      by (cases "check_cr_proof a ?i I J R prf", auto)
    note res = check_cr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding TRS_Confluence_Proof by auto
  next
    case (TRS_Non_Confluence_Proof nfs R "prf")
    from answer have ok: "isOK (check_ncr_proof a ?i I J R prf) \<and> xml_tag cp = info"
      unfolding TRS_Non_Confluence_Proof
      by (cases "check_ncr_proof a ?i I J R prf", auto)
    note res = check_ncr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding TRS_Non_Confluence_Proof by auto
  next
    case (Completion_Proof E R "prf")
    from answer have ok: "isOK (check_completion_proof a ?i I J E R prf) \<and> xml_tag cp = info"
      unfolding Completion_Proof
      by (cases "check_completion_proof a ?i I J E R prf", auto)
    note res = check_completion_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Completion_Proof by auto
  next
    case (Equational_Proof E eq "prf")
    from answer have ok: "isOK (check_equational_proof a ?i I J E eq prf) \<and> xml_tag cp = info"
      unfolding Equational_Proof by (cases "check_equational_proof a ?i I J E eq prf") auto
    note res = check_equational_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Equational_Proof by auto
  next
    case (Equational_Disproof E eq "prf")
    from answer have ok: "isOK (check_equational_disproof a ?i I J E eq prf) \<and> xml_tag cp = info"
      unfolding Equational_Disproof by (cases "check_equational_disproof a ?i I J E eq prf") auto
    note res = check_equational_disproof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Equational_Disproof by auto
  next
    case (Quasi_Reductive_Proof ctrs "prf")
    from answer have ok: "isOK (check_quasi_reductive_proof a ?i I J ctrs prf) \<and> xml_tag cp = info"
      unfolding Quasi_Reductive_Proof by (cases "check_quasi_reductive_proof a ?i I J ctrs prf") auto
    note res = check_quasi_reductive_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Quasi_Reductive_Proof by auto
  next
    case (Conditional_CR_Proof ctrs "prf")
    from answer have ok: "isOK (check_conditional_cr_proof a ?i I J ctrs prf) \<and> xml_tag cp = info"
      unfolding Conditional_CR_Proof by (cases "check_conditional_cr_proof a ?i I J ctrs prf") auto
    note res = check_conditional_cr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Conditional_CR_Proof by auto
  next
    case (Conditional_Non_CR_Proof ctrs "prf")
    from answer have ok: "isOK (check_conditional_ncr_proof a ?i I J ctrs prf) \<and> xml_tag cp = info"
      unfolding Conditional_Non_CR_Proof by (cases "check_conditional_ncr_proof a ?i I J ctrs prf") auto
    note res = check_conditional_ncr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Conditional_Non_CR_Proof by auto
  next
    case (Tree_Automata_Closed_Proof ta r "prf")
    from answer have ok: "isOK (tree_aut_trs_closed ta prf r)" "xml_tag cp = info"
      unfolding Tree_Automata_Closed_Proof by (cases "tree_aut_trs_closed ta prf r", auto)+
    from tree_aut_trs_closed[OF ok(1)] ok(2)
    show ?thesis unfolding Tree_Automata_Closed_Proof by auto
  next
    case (AC_Termination_Proof r a c "prf")
    let ?tp = "ac_tp_ops.mk K r a c"
    interpret K: ac_tp_spec K by fact
    from answer have ok: "isOK (check_ac_termination_proof L K ?i ?tp prf)" "xml_tag cp = info"
      unfolding AC_Termination_Proof by (auto split: sum.splits)
    from check_ac_termination_proof[OF L K this(1)] this(2)
    show ?thesis unfolding AC_Termination_Proof by simp
  next
    case (LTS_Termination_Proof r "prf")
    let ?tp = "lts_of' r"
    from answer have ok: "isOK (LIA_logic.check_termination (of_LIA'_lts r) prf)" "xml_tag cp = info"
      unfolding LTS_Termination_Proof by (auto split: sum.splits)
    from LIA_logic.check_termination[OF this(1)] this(2)
    show ?thesis unfolding LTS_Termination_Proof by (simp add: lts_of'_def)
  next
    case (LTS_Safety_Proof r e "prf")
    let ?tp = "lts_of r"
    from answer have ok: "isOK (LIA_logic.check_safety (of_LIA'_lts r) e prf)" "xml_tag cp = info"
      unfolding LTS_Safety_Proof by (auto split: sum.splits)
    from LIA_logic.check_safety[OF this(1)] this(2)
    show ?thesis unfolding LTS_Safety_Proof by (simp add: lts_of'_def)
  next
    case (Unknown_Proof up "prf")
    from answer have ok: "isOK (check_unknown_proof I J a ?i up prf) \<and> xml_tag cp = info"
      unfolding Unknown_Proof by (cases "check_unknown_proof I J a ?i up prf") auto
    note res = check_unknown_proof_with_assms[OF I J ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Unknown_Proof by auto
  next
    case (Unknown_Disproof up "prf")
    from answer have ok: "isOK (check_unknown_disproof I J a ?i up prf) \<and> xml_tag cp = info"
      unfolding Unknown_Disproof by (cases "check_unknown_disproof I J a ?i up prf") auto
    note res = check_unknown_disproof_with_assms[OF I J ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Unknown_Disproof by auto
  qed
qed

lemma certify_cert_problem_sound:
  assumes answer: "certify_cert_problem False cp = Certified info"
  shows "desired_property cp \<and> xml_tag cp = info"
  using certify_cert_problem_with_assms_sound[OF answer] by simp

end
end
end
