theory Ordered_Algebra
  imports "$AFP/Automatic_Refinement/Lib/Misc"
   "Quasi_Order" "~~/HOL/Lattices"
begin

section {* F-Algebras *}

locale algebra = fixes I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

fun eval :: "('f,'v) term \<Rightarrow> ('v \<Rightarrow> 'a) \<Rightarrow> 'a"
  where "eval (Var x) \<alpha> = \<alpha> x"
      | "eval (Fun f ss) \<alpha> = I f (map (\<lambda>s. eval s \<alpha>) ss)"

notation eval ("\<lbrakk>(_)\<rbrakk>")

definition subst_assign where "subst_assign \<sigma> \<alpha> x \<equiv> \<lbrakk>\<sigma> x\<rbrakk>\<alpha>"

lemma subst_assign: "\<lbrakk>s\<rbrakk>(subst_assign \<sigma> \<alpha>) = \<lbrakk>s\<cdot>\<sigma>\<rbrakk>\<alpha>"
proof(induct s)
  case (Var x) show ?case unfolding subst_assign_def by auto
next
  case (Fun f ss)
  { fix x note impI[of "x \<in> set ss", OF Fun[of x], simplified] }
  note * = map_ext[OF this]
  show ?case by (auto simp: comp_def *)
qed

lemma eval_same_vars[intro]: "(\<And>x. x \<in> vars_term t \<Longrightarrow> \<alpha> x = \<beta> x) \<Longrightarrow> \<lbrakk>t\<rbrakk> \<alpha> = \<lbrakk>t\<rbrakk> \<beta>"
  by(induct t, auto intro!:map_cong[OF refl] cong[of "I _"])

lemma eval_map_vars_term[simp]:
  "\<lbrakk>map_term id fv t\<rbrakk> \<alpha> = \<lbrakk>t\<rbrakk> (\<alpha> \<circ> fv)" by (induct t, auto simp: o_def intro: cong[of "I _"])

end

locale term_algebra = algebra "Fun :: 'f \<Rightarrow> ('f,'v) term list \<Rightarrow> ('f,'v) term"
begin

lemma interpret_is_subst[simp]: "\<lbrakk>s\<rbrakk>\<sigma> = s\<cdot>\<sigma>" by (induct s, unfold algebra.eval.simps, auto)

lemma subst_assign_is_comp[simp]: "subst_assign \<sigma> \<tau> = \<sigma> \<circ>\<^sub>s \<tau>" by (auto simp: algebra.subst_assign_def subst_compose)

end

subsection {* Quasi-ordered algebra *}

text
{* A quasi-ordered algebra is a F-algebra whose domain is quasi-ordered.
   We cannot use @{class quasi_order} as a class since later we want to use
   arbitrary order on terms which cannot be globally fixed. Local instantiation
   would alleviate the problem.
 *}

locale quasi_ordered_algebra =
  base: quasi_order less_eq less + algebra I
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

interpretation ord_syntax less_eq less.

text{* Base orderings are extended over terms in the standard way. *}

definition less_term (infix "\<sqsubset>\<^sub>t" 50) where "s \<sqsubset>\<^sub>t t \<equiv> \<forall>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>"
definition less_eq_term (infix "\<sqsubseteq>\<^sub>t" 50) where "s \<sqsubseteq>\<^sub>t t \<equiv> \<forall>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>"

abbreviation(input) greater_term (infix "\<sqsupset>\<^sub>t" 50) where "s \<sqsupset>\<^sub>t t \<equiv> t \<sqsubset>\<^sub>t s"
abbreviation(input) greater_eq_term (infix "\<sqsupseteq>\<^sub>t" 50) where "s \<sqsupseteq>\<^sub>t t \<equiv> t \<sqsubseteq>\<^sub>t s"

abbreviation Less_eq_term ("{\<sqsubseteq>\<^sub>t}") where "{\<sqsubseteq>\<^sub>t} \<equiv> {(s,t). s \<sqsubseteq>\<^sub>t t}"
abbreviation Less_term ("{\<sqsubset>\<^sub>t}") where "{\<sqsubset>\<^sub>t} \<equiv> {(s,t). s \<sqsubset>\<^sub>t t}"
abbreviation Greater_eq_term ("{\<sqsupseteq>\<^sub>t}") where "{\<sqsupseteq>\<^sub>t} \<equiv> {(s,t). s \<sqsupseteq>\<^sub>t t}"
abbreviation Greater_term ("{\<sqsupset>\<^sub>t}") where "{\<sqsupset>\<^sub>t} \<equiv> {(s,t). s \<sqsupset>\<^sub>t t}"

definition equiv_term (infix "\<sim>\<^sub>t" 50) where [iff]: "s \<sim>\<^sub>t t \<equiv> \<forall>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sim> \<lbrakk>t\<rbrakk>\<alpha>"
definition equiv_class_term ("[_]\<^sub>t") where "[s]\<^sub>t \<equiv> { t. s \<sim>\<^sub>t t }"

lemma less_termI[intro]: "(\<And>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<sqsubset>\<^sub>t t" by (auto simp: less_term_def)
lemma less_termD[dest]: "s \<sqsubset>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>t\<rbrakk>\<alpha>" by (auto simp: less_term_def)

lemma less_eq_termI[intro]: "(\<And>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<sqsubseteq>\<^sub>t t" by (auto simp: less_eq_term_def)
lemma less_eq_termD[dest]: "s \<sqsubseteq>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>t\<rbrakk>\<alpha>" by (auto simp: less_eq_term_def)

lemma equiv_termI[intro]: "(\<And>\<alpha>. \<lbrakk>s\<rbrakk>\<alpha> \<sim> \<lbrakk>t\<rbrakk>\<alpha>) \<Longrightarrow> s \<sim>\<^sub>t t" by auto
lemma equiv_termD[dest]: "s \<sim>\<^sub>t t \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sim> \<lbrakk>t\<rbrakk>\<alpha>" by auto

lemma equiv_term: "s \<sim>\<^sub>t t \<longleftrightarrow> s \<sqsubseteq>\<^sub>t t \<and> t \<sqsubseteq>\<^sub>t s" by fastforce

sublocale quasi_order "op \<sqsubseteq>\<^sub>t" "op \<sqsubset>\<^sub>t"
  apply (unfold_locales, unfold less_term_def less_eq_term_def)
  apply (auto dest: base.order_trans base.less_trans base.le_less_trans base.less_le_trans)
  done

lemma less_stable[intro]: "s \<sqsubset>\<^sub>t t \<Longrightarrow> s\<cdot>\<sigma> \<sqsubset>\<^sub>t t\<cdot>\<sigma>" by (intro less_termI, fold subst_assign, auto)
lemma lesseq_stable[intro]: "s \<sqsubseteq>\<^sub>t t \<Longrightarrow> s\<cdot>\<sigma> \<sqsubseteq>\<^sub>t t\<cdot>\<sigma>" by (intro less_eq_termI, fold subst_assign, auto)

end

locale wf_algebra = quasi_ordered_algebra less_eq less I + base: wf_order less_eq less
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool" and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
begin

interpretation ord_syntax less_eq less.

sublocale wf_order "op \<sqsubseteq>\<^sub>t :: ('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" "op \<sqsubset>\<^sub>t"
proof
  obtain v where "(v :: 'a) = v" by auto
  define \<alpha> where "\<alpha> \<equiv> \<lambda>x :: 'v. v"
  fix P s
  assume *: "\<And>s :: ('f,'v) term. (\<And>t. t \<sqsubset>\<^sub>t s \<Longrightarrow> P t) \<Longrightarrow> P s"
  then show "P s"
  proof(induct "\<lbrakk>s\<rbrakk>\<alpha>" arbitrary: s rule: base.less_induct)
    case less show ?case
    proof (rule less(2))
      fix t assume ts: "t \<sqsubset>\<^sub>t s"
      then have "\<lbrakk>t\<rbrakk>\<alpha> \<sqsubset> \<lbrakk>s\<rbrakk>\<alpha>" by auto
      from less(1)[OF this *] show "P t" by auto
    qed
  qed
qed

end


lemma(in quasi_order) weak_mono_iff_all:
  "(\<forall>a b ls rs. a \<le> b \<longrightarrow> f (ls @ a # rs) \<le> f (ls @ b # rs)) \<longleftrightarrow>
   (\<forall>as bs. length as = length bs \<longrightarrow> (\<forall>i < length as. as ! i \<le> bs ! i) \<longrightarrow> f as \<le> f bs)"
  (is "?l \<longleftrightarrow> ?r")
proof(intro iffI impI allI)
  fix as bs
  assume mono: ?l and all: "\<forall>i < length as. as ! i \<le> bs ! i" and len: "length as = length bs"
  { fix cs
    from all have "\<And>i. i < length as \<Longrightarrow> as ! i \<le> bs ! i" by auto
    with len have "f (cs @ as) \<le> f (cs @ bs)"
    proof(induct as arbitrary: bs cs)
      case Nil then have "bs = []" by auto
      with order_refl show ?case by auto
    next
      case IH: (Cons a as)
      then have "Suc (length as) = length bs" by auto
      then obtain d ds where bs: "bs = d # ds" and len: "length as = length ds" by (induct bs, auto)
      have "f ((cs @ [a]) @ as) \<le> f ((cs @ [a]) @ ds)"
        apply(rule IH(1)[OF len]) using IH(3)[unfolded bs] by force
      also have "... \<le> f ((cs @ [d]) @ ds)"
        proof-
          from IH(3)[unfolded bs, of 0] have "a \<le> d" by auto
          with mono show ?thesis by auto
        qed
      finally show ?case by (unfold bs, auto)
    qed
  }
  from this[of "[]"] show "f as \<le> f bs" by auto
next
  assume ?r
  then have *: "\<And>as bs. length as = length bs \<Longrightarrow> (\<forall>i<length as. as ! i \<le> bs ! i) \<Longrightarrow> f as \<le> f bs"
    by auto
  fix a b ls rs
  assume ab: "a \<le> b"
  show "f (ls @ a # rs) \<le> f (ls @ b # rs)"
  proof (rule *,simp, intro allI impI)
    fix i assume i: "i < length (ls @ a # rs)"
    show " (ls @ a # rs) ! i \<le> (ls @ b # rs) ! i"
    by(cases i "length ls" rule: linorder_cases, auto simp: ab nth_append)
  qed
qed

locale weak_mono_fun = base: quasi_order less_eq less
  for less_eq less +
  fixes f
  assumes mono: "\<And>ls rs a b. less_eq a b \<Longrightarrow> less_eq (f (ls @ a # rs)) (f (ls @ b # rs))"
begin

lemma mono_all:
  assumes "length as = length bs"
    and "\<And>i. i < length as \<Longrightarrow> less_eq (as ! i) (bs ! i)"
  shows "less_eq (f as) (f bs)"
proof-
  from mono
  have "\<forall>as bs. length as = length bs \<longrightarrow> (\<forall>i<length as. less_eq (as ! i) (bs ! i)) \<longrightarrow> less_eq (f as) (f bs)"
    by(subst base.weak_mono_iff_all[symmetric], auto)
  with assms show ?thesis by auto
qed

end

subsection {* Weakly monotone algebra *}

locale weak_mono_algebra = quasi_ordered_algebra less_eq less I
  for less_eq and less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  assumes weak_mono: "\<And>f ls rs a b. less_eq a b \<Longrightarrow> less_eq (I f (ls @ a # rs)) (I f (ls @ b # rs))"
begin

sublocale weak_mono_fun less_eq less "I f" by(unfold_locales, fact weak_mono)

interpretation weak_mono_term: weak_mono_fun less_eq_term less_term "Fun f"
  by (unfold_locales, intro less_eq_termI, auto intro!: weak_mono)

lemmas weak_mono_term = weak_mono_term.mono
lemmas weak_mono_term_all = weak_mono_term.mono_all

interpretation ord_syntax less_eq less.

lemma ctxt_closed_NS: "ctxt.closed {(s,t). s \<sqsupseteq>\<^sub>t t}"
 by (rule one_imp_ctxt_closed, auto intro: weak_mono_term.mono)
lemma subst_closed_S: "subst.closed {(s,t). s \<sqsupset>\<^sub>t t}" by auto
lemma subst_closed_NS: "subst.closed {(s,t). s \<sqsupseteq>\<^sub>t t}" by auto

lemma lesseq_ctxt_closed[intro]: "s \<sqsupseteq>\<^sub>t t \<Longrightarrow> C\<langle>s\<rangle> \<sqsupseteq>\<^sub>t C\<langle>t\<rangle>"
  using ctxt.closedD[OF ctxt_closed_NS] by auto

lemma mono_range: "(\<And>v. v \<in> vars_term s \<Longrightarrow> \<alpha> v \<sqsubseteq> \<beta> v) \<Longrightarrow> \<lbrakk>s\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>s\<rbrakk>\<beta>"
proof (induct s)
  case (Var v) then show ?case by auto
next
  case IH: (Fun f ss)
  { fix i assume i: "i < length ss"
    then have "vars_term (ss!i) \<subseteq> vars_term (Fun f ss)" by fastforce
    with IH(2) have *: "v \<in> vars_term (ss!i) \<Longrightarrow> \<alpha> v \<sqsubseteq> \<beta> v" for v by blast
    with i have "\<lbrakk>ss ! i\<rbrakk>\<alpha> \<sqsubseteq> \<lbrakk>ss ! i\<rbrakk>\<beta>" by(intro IH(1), auto)
  }
  then show ?case by(auto intro: mono_all)
qed

end

locale wf_weak_mono_algebra = weak_mono_algebra less_eq less I + base: wf_order less_eq less
  for less_eq less I
begin

sublocale wf_algebra..

sublocale redpair: redpair "{\<sqsupset>\<^sub>t}" "{\<sqsupseteq>\<^sub>t}"
  using SN ctxt_closed_NS subst_closed_S subst_closed_NS by unfold_locales

end


locale mono_algebra = weak_mono_algebra less_eq less I for less_eq less I +
  assumes mono: "\<And>f ls rs a b. less a b \<Longrightarrow> less (I f (ls @ a # rs)) (I f (ls @ b # rs))"
begin

lemma mono_term: "s \<sqsubset>\<^sub>t t \<Longrightarrow> Fun f (ls @ s # rs) \<sqsubset>\<^sub>t Fun f (ls @ t # rs)"
  by(intro less_termI, auto intro: mono)

lemma ctxt_closed_S: "ctxt.closed {\<sqsupset>\<^sub>t}"
  by (rule one_imp_ctxt_closed, auto intro: mono_term)

lemma less_ctxt_closed[intro]: "s \<sqsupset>\<^sub>t t \<Longrightarrow> C\<langle>s\<rangle> \<sqsupset>\<^sub>t C\<langle>t\<rangle>"
  using ctxt.closedD[OF ctxt_closed_S] by auto

end

locale wf_mono_algebra = mono_algebra less_eq less I + base: wf_order less_eq less
  for less_eq less I
begin
  sublocale wf_weak_mono_algebra..
end

section {* Models *}

locale quasi_model = weak_mono_algebra less_eq less I
  for less_eq less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  fixes R :: "('f,'v) trs"
  assumes compat: "(l,r) \<in> R \<Longrightarrow> l \<sqsupseteq>\<^sub>t r"
begin

lemma rsteps_imp_lesseq: "(s,t) \<in> (rstep R)\<^sup>* \<Longrightarrow> s \<sqsupseteq>\<^sub>t t"
proof(induct rule: rtrancl_induct)
  case base show ?case by auto
next
  case (step t u)
  then obtain C \<sigma> l r where t: "t = C\<langle>l\<cdot>\<sigma>\<rangle>" and u: "u = C\<langle>r\<cdot>\<sigma>\<rangle>" and lr: "(l,r) \<in> R" by auto
  have "t \<sqsupseteq>\<^sub>t u" by (unfold t u, auto intro: compat lr)
  also note step(3)
  finally show ?case.
qed

end

locale equimodel = weak_mono_algebra less_eq less I
  for less_eq less and I :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a" +
  fixes E :: "('f,'v) trs"
  assumes compat: "(l,r) \<in> E \<Longrightarrow> l \<sim>\<^sub>t r"
begin

sublocale quasi_model less_eq less I "E\<^sup>\<leftrightarrow>"
  by (unfold_locales, insert compat[unfolded equiv_term_def], auto)

lemma conversion_imp_equiv: assumes "(s,t) \<in> (rstep E)\<^sup>\<leftrightarrow>\<^sup>*" shows "s \<sim>\<^sub>t t"
proof-
  note * = rsteps_imp_lesseq[unfolded rstep_simps(5), folded conversion_def]
  from *[OF assms[unfolded conversion_inv[of s t]]] *[OF assms]
  show ?thesis by (auto simp: equiv_term)
qed

end

locale model = equimodel less_eq less I E + base: weak_order less_eq less for less_eq less I E
begin

lemma conversion_imp_eq: assumes "(s,t) \<in> (rstep E)\<^sup>\<leftrightarrow>\<^sup>*" shows "\<lbrakk>s\<rbrakk> = \<lbrakk>t\<rbrakk>"
  using conversion_imp_equiv[OF assms] by auto

end

subsection {* Encoding algebras *}

text {* Encoding $F$-algebra into $G$-algebra.
 Each function symbol $f \in F$ of arity $n$ is encoded as a term in $T(G,{0..<n})$. *}

locale encoding =
  fixes E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
  assumes var_domain: "i \<in> vars_term (E (f,n)) \<Longrightarrow> i < n"
begin

end


locale encoded_algebra = source: algebra I + encoding E
  for I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

definition IE :: "'f \<Rightarrow> 'a list \<Rightarrow> 'a"
  where "IE f vs = source.eval (E (f, length vs)) (nth vs)"

sublocale algebra IE.

end

locale encoded_quasi_ordered_algebra =
  base: quasi_order less_eq less + encoding E
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

sublocale encoded_algebra I E..

sublocale source: quasi_ordered_algebra less_eq less I..

sublocale quasi_ordered_algebra less_eq less IE..

end

locale encoded_wf_algebra =
  base: wf_order less_eq less + encoding E
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term"
begin

sublocale encoded_quasi_ordered_algebra less_eq less I E..

sublocale source: wf_algebra less_eq less I by (unfold_locales, fact base.less_induct)

sublocale wf_algebra less_eq less IE..

end


locale encoded_weak_mono_algebra =
  base: quasi_order less_eq less + encoding E
  for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  and I :: "'g \<Rightarrow> 'a list \<Rightarrow> 'a"
  and E :: "'f \<times> nat \<Rightarrow> ('g, nat) term" +
  assumes mono: "\<And>f ls rs a b. less_eq a b \<Longrightarrow> less_eq (I f (ls @ a # rs)) (I f (ls @ b # rs))"

begin

sublocale target: weak_mono_algebra less_eq less I by (unfold_locales, fact mono)

sublocale encoded_quasi_ordered_algebra less_eq less I E..

interpretation ord_syntax less_eq less.

sublocale weak_mono_algebra less_eq less IE
proof (unfold_locales)
  fix f ls rs a b
  { fix as bs :: "'a list"
    assume "length as = length bs"
    with var_domain[of _ f]
    have "(\<forall>i < length as. as ! i \<sqsubseteq> bs ! i) \<Longrightarrow> IE f as \<sqsubseteq> IE f bs"
      by (auto simp: IE_def intro: target.mono_range)
  }
  note * = this
  assume ab: "a \<sqsubseteq> b"
  show "IE f (ls @ a # rs) \<sqsubseteq> IE f (ls @ b # rs)"
  proof(intro * allI impI, simp)
    fix i assume "i < length (ls @ a # rs)"
    from ab show "(ls @ a # rs) ! i \<sqsubseteq> (ls @ b # rs) ! i"
    by(cases i "length ls" rule: linorder_cases, auto simp: nth_append)
  qed
qed

end

locale encoded_wf_weak_mono_algebra =
  encoded_weak_mono_algebra less_eq less I E + base: wf_order less_eq less for less_eq less I E
begin

sublocale target: encoded_wf_algebra less_eq less I E..

sublocale wf_weak_mono_algebra less_eq less IE..

end

section {* Max-polynomial Interpretations as an Instance *}

text {* As a target algebra, we introduce a $G$-algebra where $G = \{+,\times,\max\}$. *}

locale maxpoly begin

datatype sig = Fconst nat | Fsum | Fprod | Fmax

fun I where
  "I (Fconst n) = (\<lambda>x. n)"
| "I Fsum = sum_list"
| "I Fprod = prod_list"
| "I Fmax = Max \<circ> set"

sublocale wf_weak_mono_algebra less_eq less "I :: _ \<Rightarrow> nat list \<Rightarrow> nat"
proof
  fix f ls rs and a b :: nat
  assume ba: "b \<le> a"
  then show "I f (ls @ b # rs) \<le> I f (ls @ a # rs)"
  proof(cases f)
    case Fmax
    note ba
    also have "a \<le> Max (insert a (set ls \<union> set rs))" by auto
    finally show ?thesis by (auto simp: Fmax)
  qed auto
qed
(* Why do I loose these notations? *)
notation "local.less_term" (infix "\<sqsubset>\<^sub>t" 50)
notation "local.less_eq_term" (infix "\<sqsubseteq>\<^sub>t" 50)
notation "local.greater_term" (infix "\<sqsupset>\<^sub>t" 50)
notation "local.greater_eq_term" (infix "\<sqsupseteq>\<^sub>t" 50)


end

text {* Now we encode arbitrary $F$-terms into the $G$-terms. *}

locale maxpoly_interpretation = encoding E
  for E :: "'f \<times> nat \<Rightarrow> (maxpoly.sig, nat) term"
begin

sublocale target: maxpoly.

sublocale encoded_wf_weak_mono_algebra "less_eq::nat\<Rightarrow>nat\<Rightarrow>bool" less maxpoly.I E
  by (unfold_locales, fact target.mono)

text {* Now we get a reduction pair for free! *}
thm redpair.redpair_axioms

end


















end
